import pathlib

PROFILE_FILENAME = ".neosctl"
PROFILE_FILEPATH = pathlib.Path().home() / PROFILE_FILENAME


DEFAULT_PROFILE = "default"


SUCCESS_CODE = 200
REDIRECT_CODE = 300
BAD_REQUEST_CODE = 400
UNAUTHORISED_CODE = 401
